
RBProgramNodeVisitor subclass: #TNodeTranslator
	instanceVariableNames: 'tMethodClass '
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Compiler-Syntax'!

!TNodeTranslator commentStamp: 'ajh 3/24/2003 21:34' prior: 0!
I visit an abstract syntax tree rooted at RBMethodNode and generate a C-translator abstract systax tree.  This is used when translating VM Slang/Smalltalk into C.!


!RBMethodNode methodsFor: 'C translation' stamp: 'ajh 3/17/2003 01:11'!
asTranslationMethodOfClass: aClass

	^ TNodeTranslator new
		tMethodClass: aClass;
		visitNode: self! !


!TMethod methodsFor: 'initialization' stamp: 'ajh 1/21/2003 13:10'!
setSelector2: sel args: argList locals: localList block: tBlockNode primitive: aNumber
	"Initialize this method using the given information."

	selector _ sel.
	returnType _ 'int'. 	 "assume return type is int for now"
	args _ argList asOrderedCollection.
	locals _ localList asOrderedCollection.
	declarations _ Dictionary new.
	primitive _ aNumber.
	parseTree _ tBlockNode.
	labels _ OrderedCollection new.
	complete _ false.  "set to true when all possible inlining has been done"
	export _ self extractExportDirective.
	static _ self extractStaticDirective.
	self removeFinalSelfReturn.
	self recordDeclarations.
! !

!TMethod methodsFor: 'transformations' stamp: 'ajh 3/17/2003 08:31'!
removeFinalSelfReturn
	"The Smalltalk parser automatically adds the statement '^self' to the end of methods without explicit returns. This method removes such statements, since the generated code has no notion of 'self' anyway."

	| stmtList lastStmt |
	stmtList _ parseTree statements asOrderedCollection.
	stmtList isEmpty ifTrue: [^ self].
	lastStmt _ stmtList last.

	((lastStmt isReturn) and:
	 [(lastStmt expression isVariable) and:
	 ['self' = lastStmt expression name]]) ifTrue: [
		stmtList removeLast.
		parseTree setStatements: stmtList.
	].! !

!TMethod methodsFor: 'inlining' stamp: 'ajh 3/17/2003 08:40'!
statementsListsForInlining
	"Answer a collection of statement list nodes that are candidates for inlining. Currently, we cannot inline into the argument blocks of and: and or: messages."

	| stmtLists |
	stmtLists _ OrderedCollection new: 10.
	parseTree nodesDo: [ :node | 
		node isStmtList ifTrue: [ stmtLists add: node ].
	].
	parseTree nodesDo: [ :node | 
		node isSend ifTrue: [
			((node selector = #and:) or: [node selector = #or:]) ifTrue: [
				"Note: the PP 2.3 compiler produces two arg nodes for these selectors"
				stmtLists remove: node args first ifAbsent: [].
				stmtLists remove: node args last ifAbsent: [].
			].
			((node selector = #ifTrue:) or: [node selector = #ifFalse:]) ifTrue: [
				stmtLists remove: node receiver ifAbsent: [].
			].
			((node selector = #ifTrue:ifFalse:) or: [node selector = #ifFalse:ifTrue:]) ifTrue: [
				stmtLists remove: node receiver ifAbsent: [].
			].
			((node selector = #whileFalse:) or: [node selector = #whileTrue:]) ifTrue: [
				"Allow inlining if it is a [...] whileTrue/whileFalse.
				This is identified by having more than one statement in the 
				receiver block in which case the C code wouldn't work anyways"
				node receiver statements size = 1
					ifTrue:[stmtLists remove: node receiver ifAbsent: []].
			].
			(node selector = #to:do:) ifTrue: [
				stmtLists remove: node receiver ifAbsent: [].
				stmtLists remove: node args first ifAbsent: [].
			].
			(node selector = #to:by:do:) ifTrue: [
				stmtLists remove: node receiver ifAbsent: [].
				stmtLists remove: node args first ifAbsent: [].
				stmtLists remove: (node args at: 2) ifAbsent: [].
			].
		].
		node isCaseStmt ifTrue: [
			"don't inline cases"
			node cases do: [: case | stmtLists remove: case ifAbsent: [] ].
		].
	].
	^stmtLists! !


!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:29'!
acceptArrayNode: anArrayNode 

	^ self unexpected: anArrayNode! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:27'!
acceptAssignmentNode: anAssignmentNode 

	^ TAssignmentNode new
		setVariable: (self visitNode: anAssignmentNode variable)
		expression: (self visitNode: anAssignmentNode value);
		comment: anAssignmentNode comment! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:35'!
acceptBlockNode: aBlockNode 

	| node |
	node _ self visitNode: aBlockNode body.
	node setArguments: (aBlockNode arguments collect: [:arg | arg name]).
	^ node! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:38'!
acceptCascadeNode: aCascadeNode 

	^ TStmtListNode new
		setArguments: #()
		statements: (aCascadeNode messages collect: [:msg | self visitNode: msg]);
		comment: aCascadeNode comment! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:39'!
acceptDoItNode: aDoItNode 

	^ self unexpected: aDoItNode! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:40'!
acceptLiteralNode: aLiteralNode

	^ TConstantNode new setValue: aLiteralNode value! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:43'!
acceptMessageNode: aMessageNode 

	^ TSendNode new
		setSelector: aMessageNode selector
		receiver: (self visitNode: aMessageNode receiver)
		arguments: (aMessageNode arguments collect: [:arg | self visitNode: arg])! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 01:09'!
acceptMethodNode: aMethodNode 

	^ tMethodClass new
		setSelector2: aMethodNode selector
		args: aMethodNode argumentNames
		locals: aMethodNode body allDefinedVariables
		block: (self visitNode: aMethodNode body)
		primitive: aMethodNode primitiveNode num;
		comment: aMethodNode comment
! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:55'!
acceptReturnNode: aReturnNode

	^ TReturnNode new 
		setExpression: (self visitNode: aReturnNode value);
		comment: aReturnNode comment! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:58'!
acceptSequenceNode: aSequenceNode 

	| statementList |
	statementList _ OrderedCollection new.
	aSequenceNode statements do: [ :s | | newS |
		newS _ self visitNode: s.
		newS isStmtList ifTrue: [
			"inline the statement list returned when a CascadeNode is translated"
			statementList addAll: newS statements.
		] ifFalse: [
			statementList add: newS.
		].
	].
	^ TStmtListNode new
		setArguments: #()
		statements: statementList;
		comment: aSequenceNode comment! !

!TNodeTranslator methodsFor: 'visitor-double dispatching' stamp: 'ajh 3/17/2003 00:58'!
acceptVariableNode: aVariableNode

	^ TVariableNode new setName: aVariableNode name
! !

!TNodeTranslator methodsFor: 'initialize' stamp: 'ajh 3/17/2003 01:09'!
tMethodClass: aTMethodClass

	tMethodClass _ aTMethodClass! !


!TestCodeGenerator methodsFor: 'coercing' stamp: 'ajh 2/7/2003 21:27'!
ccgTVarBlock: anInteger

	| ctxtString |
	ctxtString _ Preferences compileBlocksAsClosures
		ifTrue: ['thisContext']  "thisContext refers to captured context not eval context"
		ifFalse: ['(thisContext tempAt: 1)'].  "thisContext refers to eval context"
	^ [:expr | ctxtString, ' tempAt: ', anInteger asString, ' put: (', expr, ')']! !


!TestTMethod methodsFor: 'initializing' stamp: 'ajh 1/21/2003 13:11'!
setSelector2: sel args: argList locals: localList block: tBlockNode primitive: aNumber
	"Initialize this method using the given information."

	selector _ sel.
	returnType _ 'int'. 	 "assume return type is int for now"
	args _ argList asOrderedCollection.
	locals _ localList asOrderedCollection.
	declarations _ Dictionary new.
	primitive _ aNumber.
	parseTree _ tBlockNode.
	labels _ OrderedCollection new.
	complete _ false.  "set to true when all possible inlining has been done"
	export _ self extractExportDirective.
	static _ self extractStaticDirective.
	isPrimitive _ false.  "set to true only if you find a primtive direction."
	suppressingFailureGuards _ self extractSuppressFailureGuardDirective.
	self recordDeclarations.
	self extractPrimitiveDirectives.
! !
